// The following code runs speed benchmarks for the different modules
// The first inference speed for each module is discarded due to lazy initialization of models

#include <chrono>
#include <iostream>
#include <string>
#include <vector>
#include <cstdlib> 

#include "tf_sdk.h"


typedef std::chrono::high_resolution_clock Clock;

void benchmarkFaceRecognitionLite(const std::string& license, unsigned int numIterations = 200);
void benchmarkFaceRecognitionFull(const std::string& license, unsigned int numIterations = 30);
void benchmarkObjectDetection(const std::string& license, unsigned int numIterations = 100);
void benchmarkFaceLandmarkDetection(const std::string& license, unsigned int numIterations = 100);
void benchmarkMaskDetection(const std::string& license, unsigned int numIterations = 1000);

int main() {
    // TODO: Replace with your license
    const std::string license = "${LICENSE_CODE}";

    std::cout << "Running speed benchmarks with 1280x720 image\n";

    benchmarkFaceRecognitionLite(license);
    benchmarkFaceRecognitionFull(license);
    benchmarkObjectDetection(license);
    benchmarkFaceLandmarkDetection(license);
    benchmarkMaskDetection(license);

    return 0;
}

void benchmarkFaceRecognitionLite(const std::string& license, unsigned int numIterations) {
    // Initialize the SDK
    Trueface::ConfigurationOptions options;
    options.frModel = Trueface::FacialRecognitionModel::LITE;

    Trueface::SDK tfSdk(options);
    bool valid = tfSdk.setLicense(license);

    if (!valid) {
        std::cout << "Error: the provided license is invalid." << std::endl;
        exit (EXIT_FAILURE);
    }

    // Load the image
    Trueface::ErrorCode errorCode = tfSdk.setImage("../images/headshot.jpg");
    if (errorCode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Error: could not load the image" << std::endl;
        return;
    }

    // Discard first inference speed
    Trueface::Faceprint faceprint;
    bool foundFace;
    tfSdk.getLargestFaceFeatureVector(faceprint, foundFace);

    // Obtain the aligned chip
    Trueface::FaceBoxAndLandmarks faceBoxAndLandmarks;
    bool found;
    tfSdk.detectLargestFace(faceBoxAndLandmarks, found);

    uint8_t faceBuffer[112*112*3];
    tfSdk.extractAlignedFace(faceBoxAndLandmarks, faceBuffer);

    // Time the creation of the feature vector
    auto t1 = Clock::now();
    for (size_t i = 0; i < numIterations; ++i) {
        tfSdk.getFaceFeatureVector(faceBuffer, faceprint);
    }
    auto t2 = Clock::now();
    double totalTime = std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count();

    std::cout << "Average time face recognition lite: " << totalTime / numIterations
              << " ms | " << numIterations << " iterations" << std::endl;

}

void benchmarkFaceRecognitionFull(const std::string& license, unsigned int numIterations) {
    // Initialize the SDK
    Trueface::ConfigurationOptions options;
    options.frModel = Trueface::FacialRecognitionModel::FULL;

    Trueface::SDK tfSdk(options);
    bool valid = tfSdk.setLicense(license);

    if (!valid) {
        std::cout << "Error: the provided license is invalid." << std::endl;
        exit (EXIT_FAILURE);
    }

    // Load the image
    Trueface::ErrorCode errorCode = tfSdk.setImage("../images/headshot.jpg");
    if (errorCode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Error: could not load the image" << std::endl;
        return;
    }

    // Discard first inference speed
    Trueface::Faceprint faceprint;
    bool foundFace;
    errorCode = tfSdk.getLargestFaceFeatureVector(faceprint, foundFace);

    // Obtain the aligned chip
    Trueface::FaceBoxAndLandmarks faceBoxAndLandmarks;
    bool found;
    tfSdk.detectLargestFace(faceBoxAndLandmarks, found);

    uint8_t faceBuffer[112*112*3];
    tfSdk.extractAlignedFace(faceBoxAndLandmarks, faceBuffer);

    if (errorCode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Error: Could not benchmark full model, SDK not built with full model support" << std::endl;
        return;
    }

    // Time the creation of the feature vector
    auto t1 = Clock::now();
    for (size_t i = 0; i < numIterations; ++i) {
        tfSdk.getFaceFeatureVector(faceBuffer, faceprint);
    }
    auto t2 = Clock::now();
    double totalTime = std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count();

    std::cout << "Average time face recognition full: " << totalTime / numIterations
    << " ms | " << numIterations << " iterations" << std::endl;
}

void benchmarkObjectDetection(const std::string& license, unsigned int numIterations) {
    // Initialize the SDK with the fast object detection model
    Trueface::ConfigurationOptions options;
    options.objModel = Trueface::ObjectDetectionModel::FAST;

    Trueface::SDK tfSdk(options);
    bool valid = tfSdk.setLicense(license);

    if (!valid) {
        std::cout << "Error: the provided license is invalid." << std::endl;
        exit (EXIT_FAILURE);
    }

    // Load the image
    Trueface::ErrorCode errorCode = tfSdk.setImage("../images/bike.jpg");
    if (errorCode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Error: could not load the image" << std::endl;
        return;
    }

    // Discard first inference speed
    std::vector<Trueface::BoundingBox> boundingBoxes;
    tfSdk.detectObjects(boundingBoxes);

    if (boundingBoxes.empty()) {
        std::cout << "Unable to detect objects\n";
        return;
    }

    // Time the creation of the feature vector
    auto t1 = Clock::now();
    for (size_t i = 0; i < numIterations; ++i) {
        tfSdk.detectObjects(boundingBoxes);

    }
    auto t2 = Clock::now();
    double totalTime = std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count();

    std::cout << "Average time object detection (fast mode): " << totalTime / numIterations
              << " ms | " << numIterations << " iterations" << std::endl;

}

void benchmarkFaceLandmarkDetection(const std::string& license, unsigned int numIterations) {
    // Initialize the SDK
    Trueface::ConfigurationOptions options;
    options.smallestFaceHeight = 40;
    Trueface::SDK tfSdk(options);
    bool valid = tfSdk.setLicense(license);

    if (!valid) {
        std::cout << "Error: the provided license is invalid." << std::endl;
        exit (EXIT_FAILURE);
    }

    // Load the image
    Trueface::ErrorCode errorCode = tfSdk.setImage("../images/headshot.jpg");
    if (errorCode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Error: could not load the image" << std::endl;
        return;
    }

    // Discard first inference speed
    Trueface::FaceBoxAndLandmarks faceBoxAndLandmarks;
    bool found = false;
    tfSdk.detectLargestFace(faceBoxAndLandmarks, found);
    if (!found) {
        std::cout << "Unable to find face in image\n";
        return;
    }

    // Time the creation of the feature vector
    auto t1 = Clock::now();
    for (size_t i = 0; i < numIterations; ++i) {
        tfSdk.detectLargestFace(faceBoxAndLandmarks, found);
    }
    auto t2 = Clock::now();
    double totalTime = std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count();

    std::cout << "Average time face and landmark detection: " << totalTime / numIterations
              << " ms | " << numIterations << " iterations" << std::endl;

}

void benchmarkMaskDetection(const std::string& license, unsigned int numIterations)
{
    // Initialize sdk
    Trueface::SDK tfSdk;

    bool valid = tfSdk.setLicense(license);

    if (!valid) {
        std::cout << "Error: the provided license is invalid." << std::endl;
        exit (EXIT_FAILURE);
    }

    // Load the image
    Trueface::ErrorCode errorCode = tfSdk.setImage("../images/headshot.jpg");
    if (errorCode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Error: could not load the image" << std::endl;
        return;
    }

    Trueface::FaceBoxAndLandmarks faceBoxAndLandmarks;
    bool found;
    tfSdk.detectLargestFace(faceBoxAndLandmarks, found);
    if (!found)
    {
        std::cout<< "Unable to find face in image." << std::endl;
    }

    // Discard first inference
    float maskScore;
    Trueface::MaskLabel maskLabel;
    tfSdk.detectMask(faceBoxAndLandmarks, maskLabel, maskScore);

    // Time the creation of the feature vector
    auto t1 = Clock::now();
    for (size_t i = 0; i < numIterations; ++i) {
        tfSdk.detectMask(faceBoxAndLandmarks, maskLabel, maskScore);
    }
    auto t2 = Clock::now();
    double totalTime = std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count();

    std::cout << "Average time mask detection: " << totalTime / numIterations
                << " ms | " << numIterations << " iterations" << std::endl;
}